@@ -28,10 +28,14 @@
 
 #include <nv/target> // __CUDA_MINIMUM_ARCH__ and friends
 
+#include <cuda/std/__cccl/prologue.h>
+
 _LIBCUDACXX_BEGIN_NAMESPACE_CUDA_PTX
 
 #include <cuda/__ptx/instructions/generated/tcgen05_alloc.h>
 
 _LIBCUDACXX_END_NAMESPACE_CUDA_PTX
 
+#include <cuda/std/__cccl/epilogue.h>
+
 #endif // _CUDA_PTX_TCGEN05_ALLOC_H_
