@@ -1,9 +1,12 @@
 """Module for RBF interpolation."""
+from __future__ import annotations
+
 import math
 import warnings
 from itertools import combinations_with_replacement
 
 import cupy as cp
+from cupyx.scipy.spatial import KDTree
 
 
 # Define the kernel functions.
@@ -652,7 +655,6 @@ class RBFInterpolator:
         if neighbors is None:
             nobs = ny
         else:
-            raise NotImplementedError("neighbors is not implemented yet")
             # Make sure the number of nearest neighbors used for interpolation
             # does not exceed the number of observations.
             neighbors = int(min(neighbors, ny))
@@ -679,8 +681,7 @@ class RBFInterpolator:
             self._coeffs = coeffs
 
         else:
-            raise NotImplementedError
-            # self._tree = KDTree(y)
+            self._tree = KDTree(y)
 
         self.y = y
         self.d = d
@@ -719,7 +720,7 @@ class RBFInterpolator:
         nnei = len(y)
 
         # in each chunk we consume the same space we already occupy
-        chunksize = memory_budget // ((self.powers.shape[0] + nnei)) + 1
+        chunksize = memory_budget // (self.powers.shape[0] + nnei) + 1
         if chunksize <= nx:
             out = cp.empty((nx, self.d.shape[1]), dtype=float)
             for i in range(0, nx, chunksize):
@@ -783,7 +784,51 @@ class RBFInterpolator:
                 self._scale,
                 self._coeffs, memory_budget=memory_budget)
         else:
-            raise NotImplementedError    # XXX: needs KDTree
+            # Get the indices of the k nearest observation points to each
+            # evaluation point.
+            _, yindices = self._tree.query(x, self.neighbors)
+
+            if self.neighbors == 1:
+                # `KDTree` squeezes the output when neighbors=1.
+                yindices = yindices[:, None]
+
+            # Multiple evaluation points may have the same neighborhood of
+            # observation points. Make the neighborhoods unique so that we only
+            # compute the interpolation coefficients once for each
+            # neighborhood.
+            yindices = cp.sort(yindices, axis=1)
+            yindices, inv = cp.unique(yindices, return_inverse=True, axis=0)
+            # `inv` tells us which neighborhood will be used by each evaluation
+            # point. Now we find which evaluation points will be using each
+            # neighborhood.
+            xindices = [[] for _ in range(len(yindices))]
+            for i, j in enumerate(inv.tolist()):
+                xindices[j].append(i)
+
+            out = cp.empty((nx, self.d.shape[1]), dtype=float)
+            for xidx, yidx in zip(xindices, yindices):
+                # `yidx` are the indices of the observations in this
+                # neighborhood. `xidx` are the indices of the evaluation points
+                # that are using this neighborhood.
+                xnbr = x[xidx]
+                ynbr = self.y[yidx]
+                dnbr = self.d[yidx]
+                snbr = self.smoothing[yidx]
+                shift, scale, coeffs = _build_and_solve_system(
+                    ynbr,
+                    dnbr,
+                    snbr,
+                    self.kernel,
+                    self.epsilon,
+                    self.powers,
+                )
+                out[xidx] = self._chunk_evaluator(
+                    xnbr,
+                    ynbr,
+                    shift,
+                    scale,
+                    coeffs,
+                    memory_budget=memory_budget)
 
         out = out.view(self.d_dtype)
         out = out.reshape((nx, ) + self.d_shape)
